# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from    os       import  listdir
from    os.path  import  basename, abspath, dirname, isfile, isdir, dirname
from    os.path  import  join                        as  pjoin
from    time     import  time
from    math     import  prod
from    sys      import  path                        as  sys_path
import  torch 
# no random seed is needed for now

# ------------------------------------------------------------------------
#                  Root Directory (interp + ml)
# ------------------------------------------------------------------------

root_dir  =  dirname(dirname(dirname(abspath(__file__))))

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------
sys_path.append(dirname(dirname(abspath(__file__))))
from  core.io_access        import  data_dirs
from  core.utils            import  import_json, as_namedtuple, lmap, lfilter, pop1, reader
sys_path.pop()

# ------------------------------------------------------------------------
#                  Interpolation Output Directory
# ------------------------------------------------------------------------

interp_output_dir  =  data_dirs.output_dir

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

# utilities
almost         =  lambda a, b   :  torch.abs(a-b).max().item() < 1e-10
linspace       =  lambda a, b, n:  torch.linspace(a, b, n, dtype = torch.double)

listdir_full   =  lambda x  : sorted([pjoin(x,y) for y in listdir(x)])

def find_once(tag , A):
    # find element in A[i] that contains "tag"
    n = 0  # number of repetitions
    for x in A:
        if tag in x:
            result  = x            # record line found
            n      += x.count(tag) # count number of matches
    assert n == 1                  # assert only one match was found
    return result

# ------------------------------------------------------------------------
#                  Import Case
# ------------------------------------------------------------------------

def Import_Case(fname_case):
    
    fname_case  =  pjoin( data_dirs.case_input_dir ,
                          basename(fname_case)     )
    
    return as_namedtuple( import_json(fname_case), 'case_json' )

# ------------------------------------------------------------------------
#                  Load Single Tecplot File
# ------------------------------------------------------------------------
#     (a staggered DNS grid is described by multiple files)

class Load_Single_Tec:
    def __init__(self, fname):
        
        with open(fname,'r') as f:
            # read header, zone
            getline   =  lambda: next(iter(f)).rstrip('\n')
            header    =  getline()
            zone      =  getline()
            
            # read matrix
            A         =  []
            for line in f:
                A.append(lmap(float,
                              filter(None,line.rstrip('\n').split(' '))
                             ))
        
        # define order
        order  =  header.split('=')[1].replace('"',' ').replace(',',' ')
        order  =  lfilter(None,order.split(' '))
        
        # convert to torch.tensor
        A      =  torch.tensor(A, dtype = torch.double)
        
        # register all variables
        self.fname   =  fname
        self.header  =  header
        self.zone    =  zone
        self.order   =  order
        self.data    =  { key:A[:,i] for i,key in enumerate(order) }
        
        self.make_all_2d()
    
    def make_all_2d(self):
        
        inds_2d = self.find_inds_2d(self.data['X'],
                                    self.data['Z'])
        
        for k,v in self.data.items():
            self.data[k] = v[inds_2d]
    
    @staticmethod
    def find_inds_2d(x, z):
        
        # mk_inds:
        def mk_inds(a, a1):
            to_ind  =  {val.item():i for i,val in enumerate(a1)}
            return  [to_ind[val.item()]  for val in  a]
        
        # 1d distributions
        x1   =  torch.unique(x)
        z1   =  torch.unique(z)
        
        # shape 2d
        shape2 = len(x1), len(z1)
        
        # assert shapes are consistent
        #     (floating point errors in "unique" would be detected here)
        assert prod(shape2) == len(x) == len(z)
        
        # build index maps
        #     (convert xz positions into indexes)
        #     e.g., [0.5 ,1.5, 0.5, ... ] -> [0, 1, 0,...]
        ix  =  mk_inds(x, x1)
        iz  =  mk_inds(z, z1)
        
        # define final array
        #     inds = [[0,1,5,6],[9,3,2]]
        #       (this indicates how elements should be regrouped to create 2d array)
        
        inds = torch.zeros( shape2 , dtype = torch.long )
        
        for k in range(len(x)):
            # inds[a,b] goes at position (ix[k],iz[k])
            inds[ix[k],iz[k]] = k
        
        return inds
        

# ------------------------------------------------------------------------
#                  Load All Tecplot Data (for one DNS case)
# ------------------------------------------------------------------------
#     (this class is intended to be a data-holder attribute for
#      ML_Data_Single_Case)

class Load_All_Tec:
    def __init__(self, subfolder, side, qmode, fprint, apply_check = True):
        
        self.fprint           =  fprint
        self.subfolder        =  subfolder
        self.full_data_folder =  pjoin( interp_output_dir, self.subfolder )
        
        self.side             =  side
        
        self.load_data ()
        if apply_check:
            self.check_data()
        self.__mk_internal_vars()
        
        if   qmode == 'Fx':
            self.Fx  =  self.__Fx
        
        elif qmode == 'Qdot':
            self.Qdot  =  self.__Qdot
        
        elif qmode == 'Cf_classic':
            self.Cf_classic = self.__Cf_classic
        
        elif qmode == 'St_classic':
            self.St_classic = self.__St_classic
        
        elif qmode == 'Nu_classic':
            self.Nu_classic = self.__Nu_classic
        
        else:
            raise Exception(f'Unrecognized (qmode = {qmode}) (expected qmode in [Fx, Qdot, Cf_classic, St_classic, Nu_classic])')
    
    def load_data(self):
        
        for var in 'UVWPT':
            fname = pjoin( self.full_data_folder, self.subfolder + f'_quadint_interp2center_{self.side}_{var}.tec' )
            setattr(self, f'tec_{var}', Load_Single_Tec(fname))
    
    def check_data(self):
        x  =  self.x_2d  =  self.tec_P.data['X']
        y                =  self.tec_P.data['Y']
        z  =  self.z_2d  =  self.tec_P.data['Z']
        h =   self.h_2d  =  self.tec_P.data['H']
        
        for key in 'UVWPT':
            data  =  getattr(self, f'tec_{key}').data
            # print(key,"torch.abs(y-data['Y']).max().item()",torch.abs(y-data['Y']).max().item())
            assert all( almost(arr,data[key]) for arr,key in ((x,'X'),(z,'Z'),(h,'H'),(y,'Y')) )
    
    def __load_bulk_data(self):
        fname = pop1(lfilter(lambda x: basename(x)[-11:]=='_DUplus.dat',
                             listdir_full(self.full_data_folder)       ))
        
        fname_rough_surf = pop1(lfilter(lambda x: basename(x)=='rough_surf_define.py',
                                listdir_full(self.full_data_folder)                  ))
        
        A     = reader(fname)
        
        def get_val(tag,f=float):
            n = 0
            for x in A:
                if tag in x:
                    assert ':' in x
                    result  = f(x.split(':')[1])
                    n      += x.count(tag)
            assert n == 1
            return result
        
        self.props                              =  {}
        self.props['Ret']   = Ret               =  get_val(' Ret ')
        self.props['L_xyz'] = __Lx, Ly, __Lz    =  get_val(' L_xyz ',f=lambda x: lmap(float,x.replace('[','').replace(']','').split(',')))
        self.props['N_xyz'] = __Nx,  _, __Nz    =  get_val(' N_xyz ',f=lambda x: lmap(int  ,x.replace('[','').replace(']','').split(',')))
        self.props['Ub']                        =  get_val(' Bulk Velocity')
        self.props['Tb']                        =  get_val(' Bulk Temperature')
        self.props['Pr']                        =  get_val(' Pr ')
        self.props['dpdx']                      =  1.
        self.props['source_q']                  =  1.
        self.props['rho']                       =  1.
        self.props['cp']                        =  1.
        self.props['Tw']                        =  0.

        self.props['nreps_x']                   =  get_val(' nreps_x ')
        self.props['nreps_z']                   =  get_val(' nreps_z ')
        
        self.props['__A_elem']                  =  (__Lx * __Lz)/(__Nx * __Nz)
        
        self.props['nu']   =  nu                =  1./Ret
        self.props['Reb']                       =  self.props['Ub'] * Ly / nu

        self.props['code_height_function']      =  reader(fname_rough_surf)
        self.props['fname_rough_surf']          =  fname_rough_surf

        self.fprint('    -------------------------------------------------')
        self.fprint('                    Reading Data File')
        self.fprint('    -------------------------------------------------')
        
        self.fprint(f'        fname  = {fname}')
        for k,v in self.props.items():
            self.fprint(f'        {k:5s}  = {v}')
        self.props['WARNING'] = "        " + f"WARNING: the following constant properties have been assumed (rho = {self.props['rho']}) (cp = {self.props['cp']}) (Tw = {self.props['Tw']}) (dpdx = {self.props['dpdx']}) (source_q = {self.props['source_q']})".replace(' ','_')
        self.fprint(self.props['WARNING'])
        
    def __mk_internal_vars(self):
        self.__mk_Fx()
        self.__mk_Qdot()
        self.__load_bulk_data()
        self.__mk_Cf_St_Nu_classic()
    
    def __mk_Fx(self):
        
        data_U  =  self.tec_U.data
        data_V  =  self.tec_V.data
        data_W  =  self.tec_W.data
        data_P  =  self.tec_P.data
        
        self.__Fx    =  ( data_U["inv_Ret_2.0_dUdx_nx"] + data_U["inv_Ret_1.0_dUdy_ny"] + data_U["inv_Ret_1.0_dUdz_nz"] + 
                                                          data_V["inv_Ret_1.0_dVdx_ny"] + 
                                                                                          data_W["inv_Ret_1.0_dWdx_nz"] + 
                          data_P["minus_P_nx"]          )
    
    def __mk_Qdot(self):
        
        data_T       =  self.tec_T.data
        self.__Qdot  =  ( data_T["inv_Pet_dTdx_nx"] +
                          data_T["inv_Pet_dTdy_ny"] +
                          data_T["inv_Pet_dTdz_nz"] )
    
    def __mk_Cf_St_Nu_classic(self):
        
        rho     =  self.props['rho']
        Ub      =  self.props['Ub']
        cp      =  self.props['cp']
        Tw      =  self.props['Tw']
        Tb      =  self.props['Tb']
        Pr      =  self.props['Pr']
        Reb     =  self.props['Reb']
        A_elem  =  self.props['__A_elem']
        
        assert (hasattr(self.__Fx  ,'shape') and
                hasattr(self.__Qdot,'shape') and
                (type(rho) ==
                 type(Ub ) == 
                 type(cp ) == 
                 type(Tw ) == 
                 type(Tb ) == 
                 type(Pr ) == 
                 type(Reb) ==  float)
               )
        
        self.__Cf_classic  =  2*self.__Fx   / A_elem / (rho*(Ub**2))
        self.__St_classic  =    self.__Qdot / A_elem / (rho*Ub*cp*(Tb-Tw))
        self.__Nu_classic  =  self.__St_classic * Reb * Pr           # remember that Reb is defined with Ly above
        
        def avg(tag, all_x, a, b):
            x  =  float(torch.mean(all_x))
            print(tag, f'{x:.8f}')
            if not (0.5*min(a,b)) < x < (1.5*max(a,b)):
                print(f"WARNING: (tag,x,a,b: {(tag,x,a,b)})")
            # assert (0.5*min(a,b)) < x < (1.5*max(a,b)), ('tag,x,a,b: ',tag,x,a,b)
        
        avg('__Cf_classic', self.__Cf_classic,  0.0150 ,  0.0175 )
        avg('__Nu_classic', self.__Nu_classic, 28.02   , 29.66   )
        avg('__St_classic', self.__St_classic,  0.0068 ,  0.0077 )
    
    def _get_full_export(self):
        result = {}
        for var_ in 'UVWPT':
            data = getattr(self, f"tec_{var_}").data
            for key in data:
                if not key in result:
                    result[key] = data[key]
        result.update(dict(Cf_classic  =  self.__Cf_classic,
                           Fx          =  self.__Fx        ,
                           Qdot        =  self.__Qdot      ,
                           Nu_classic  =  self.__Nu_classic))
        
        props = {}
        for key,val in self.props.items():
            if hasattr(val,'shape'):
                props[key] = val.tolist()
            else:
                props[key] = val
        
        return {key: val.tolist()  for key,val in  result.items()}, self.full_data_folder, props
