# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  torch
from    time     import  time
from    re       import  search  as  re_search
from    os.path  import  basename, isfile
from    os.path  import  join  as  pjoin

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.io_access  import  data_dirs
from  core.utils      import  lmap, lfilter, spacejoin, get_ram_usage_gb, format_dt, as_namedtuple, as_list

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

clone             =  torch.clone
zeros_like        =  torch.zeros_like
minimum           =  torch.minimum
maximum           =  torch.maximum

spacesplit        =  lambda s: lfilter(None, s.split(' ') )
as_item           =  lambda x: x.item()

sign              =  lambda x:  -1 if x<0 else 1

def append_dict(D, key, val):
    if not key in D:
        D[key] = [val]
    else:
        D[key].append(val)

def writer(fname, header, *arrs):
    
    with open(fname,'w') as f:
        write = f.write
        
        for x in header:
            write(x + '\n')
        
        for row in zip(*arrs):
           write( ' '.join(map(str, row)) + '\n' )

def fsum(A,B):
    for i,b in enumerate(B):
        A[i] += b

def fdiv(A,B):
    
    for i,b in enumerate(B):
        A[i] /= max(1e-10,abs(b)) * sign(b)


# ------------------------------------------------------------------------
#                  Gather Destination Addresses
# ------------------------------------------------------------------------

def Gather_Addresses(all_grids, order):
    
    D = {}
    
    for field in 'uvwPT':
        
        ind_tec  =   order.index(field)
        grid     =  all_grids(field.upper())
        
        for side in ('top','bottom'):
            
            arr_i  =  getattr(grid, f'i_{side}'   )
            arr_v  =  getattr(grid, f'vals_{side}')
            
            assert arr_i.shape == arr_v.shape
            
            for     k in range(grid.shape2[0]):
                for j in range(grid.shape2[1]):
                
                    for pos, i in enumerate(lmap(as_item, arr_i[k,j])):
                        
                        # append_dict(D, key     , val               )
                        append_dict  (D, (k,i,j) , [ind_tec, arr_v, pos] )
    
    return D


# ------------------------------------------------------------------------
#                  Data Reader
# ------------------------------------------------------------------------

class Data_Reader:
    def __init__(self, all_grids, args, fprint):
        
        self.args          =  args
        self.tecname       =  args.tecname
        
        self.reader(all_grids, fprint)
    
    def reader(self, all_grids, fprint):
        
        # define basic PyTorch conversion
        tensor      =  torch.tensor
        double      =  torch.double
        
        # make paths to unzipped_dns
        unzipped_path  =  self.tecname
        
        # assert that unzipped_dns exists
        assert  isfile(unzipped_path)
        
        # build fname
        fname          =  unzipped_path
        
        
        # -------------------- variables to calculate velocity profile ------------------
        # define dictionary with profile information
        info_prof  =  {}
        
        # define empty array with the same shape as grid.y1
        mk_zeros  =  lambda arr: [ 0. for _ in arr ]
        
        # register values for (U,T) (profile variables)
        info_prof[f'profile_u'    ]  =  profile_u      =  mk_zeros(all_grids( 'U' ).y1)
        info_prof[f'profile_u_vol']  =  profile_u_vol  =  mk_zeros(all_grids( 'U' ).y1)
        info_prof[f'profile_u_y'  ]  =                    as_list (all_grids( 'U' ).y1,float)
        
        info_prof[f'profile_T'    ]  =  profile_T      =  mk_zeros(all_grids( 'T' ).y1)
        info_prof[f'profile_T_vol']  =  profile_T_vol  =  mk_zeros(all_grids( 'T' ).y1)
        info_prof[f'profile_T_y'  ]  =                    as_list (all_grids( 'T' ).y1,float)
        
        # add bulk values for all other variables (except P)
        info_prof['bulk_uvwT']  =  bulk_uvwT  =  [0., 0., 0., 0.]
        info_prof['bulk_vols']  =  bulk_vols  =  [0., 0., 0., 0.]
    
        # assert that long file will be read
        assert fname == unzipped_path
        
        get_vols_uvwT  =  lmap( lambda c: all_grids(c).get_vol , 'UVWT' )
        # --------------------------------------------------------------------------------------
        
        # get Ret and Pr
        self.Ret = self.args.Ret
        self.Pr  = self.args.Pr
        
        # track elapsed time
        t0           =  time()
        total_lines  =  self.args.total_lines
        print_freq   =  total_lines//50
        
        # open file
        with open(fname,'r') as f:
            header = next(iter(f)).rstrip('\n')
            zone   = next(iter(f)).rstrip('\n')
            
            assert (header == self.args.ref_header) and (zone == self.args.ref_zone)
            
            order     =  self.get_order(header)
            
            inds_kij   =  lmap(lambda c: order.index(c), 'KIJ' )
            inds_uvwT  =  lmap(lambda c: order.index(c), 'uvwT')
            
            # Print header
            fprint('\n--------------------------------------------------\n    Reading DNS Data\n--------------------------------------------------\n')
            fprint(f'    Full   DNS Path   : {unzipped_path}')
            fprint(f'    Chosen DNS (Read) : {fname}')
            fprint(f'    Ret               : {self.Ret}')
            fprint(f'    Pr                : {self.Pr }')
            fprint(f'    Variable Order    : {spacejoin(order)}')
            fprint(f'    Initial RAM Usage : {get_ram_usage_gb():.2f} Gb\n')
            
            addresses_kij   =  Gather_Addresses(all_grids, order)
            
            for iread, line in enumerate(f,1):
                
                if not (iread % print_freq):
                    print(f'Progress: {round(iread/total_lines*100,1):.1f} % (Elapsed Time = {format_dt(time()-t0)})')
                
                values         =  spacesplit(line.rstrip('\n'))
                
                k,i,j  =  kij  =  tuple( int(values[ii]) for ii in inds_kij )
                
                if kij in addresses_kij:
                    for  ind_tec, arr, pos  in  addresses_kij[kij]:
                        arr[k,j,pos] = float(values[ind_tec])
                
                # ------------------ calc_profile ------------------
                
                vols_             =  [ get_vol(k,i,j)        for get_vol  in  get_vols_uvwT         ]
                uvwT_volWeighted  =  [ float(values[ii])*vol for ii,vol   in  zip(inds_uvwT,vols_)  ]
                
                # sum values in-place (for the first argument)
                fsum(bulk_vols , vols_    )
                fsum(bulk_uvwT , uvwT_volWeighted)
                
                profile_u    [i]  +=  uvwT_volWeighted [ 0]
                profile_u_vol[i]  +=  vols_            [ 0]
                
                profile_T    [i]  +=  uvwT_volWeighted [-1]
                profile_T_vol[i]  +=  vols_            [-1]
                # ----------------------------------------------------
        
        # ------------------ calc_profile ------------------
        # vol. division occurs in-place (for the first argument)
        fdiv(profile_u, profile_u_vol)
        fdiv(profile_T, profile_T_vol)
        fdiv(bulk_uvwT, bulk_vols    )
        
        # save t_sweep
        info_prof['t_sweep']  =  self.args.Lx / max(1e-10, bulk_uvwT[0])
        
        # register info_prof in object
        self.info_prof        =  as_namedtuple(info_prof, 'calc_profile_info')
        # ----------------------------------------------------
        
        fprint( '    Done!\n')
        fprint(f'    DNS Reading Time: {format_dt(time()-t0)}')
        fprint(f'    Final RAM Usage : {get_ram_usage_gb():.2f} Gb\n')
    
    def get_order(self, header):
        
        order  =  header.split('=')[1].replace('"','').replace(' ','').split(',')
        
        # assert (','.join(order)) == 'K0,I0,J0,U,V,W,P,T'
        return order