# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from    os.path  import  isdir, basename, isfile, dirname
from    os       import  mkdir
from    os.path  import  join  as  pjoin
from    time     import  time
from    math     import  prod
from    sys      import  path  as  sys_path

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from core.utils                     import  as_namedtuple, tstamp, format_dt, iformat, qformat, lmap, as_list, reader
from core.io_access                 import  data_dirs, Mk_Case, writer
from core.mk_grid                   import  All_Grids
from core.data_reader               import  Data_Reader
from core.io_access                 import  FPrint
from core.integration_driver        import  Integration_Driver
from core.summary_tec               import  SummaryBalances
from plot_du_plus .plotter_dutu     import  Plotter_DUTU_Plus

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

output_dir        =  data_dirs.output_dir
unzipped_dns_dir  =  data_dirs.unzipped_dns_dir

def get_height_function(fname):
    assert basename(fname) == 'rough_surf_define.py'
    sys_path.append(dirname(fname))
    from rough_surf_define import Height_Function
    sys_path.pop()
    return Height_Function()

# ------------------------------------------------------------------------
#                  Main Routine
# ------------------------------------------------------------------------

class Engine:
    def __init__(self, args):
        
        # args is generated by scheduler.py
        
        t0  =  time()
        
        # register arguments, height function, caseinfo
        self.args             =  args
        
        self.case    =  case  =  Mk_Case(args.folder_path)
        
        self.sheight  =  get_height_function(case.surf_define)

        # initialize outpath, fprint
        self.ini_outpath()
        self.ini_fprint()
        fprint = self.fprint

        writer(pjoin(dirname(fprint.fname), basename(case.surf_define)),
               reader(case.surf_define)                                )
        
        # print header
        fprint( '\n--------------------------------------------------\n    Processing Case\n--------------------------------------------------\n')
        fprint(f'    Folder run  : {args.folder_path }')
        fprint(f'    Surface file: {case.surf_define }')
        
        fprint(f'\nInput Arguments:')
        for k,v in args._asdict().items():
            fprint(f'    {k:15}: {v}')
        
        fprint(f'\nCase Data (json):')
        for k,v in case._asdict().items():
            fprint(f'    {k:15}: {v}')
        
        # fprint(f'\nSurface Data (json):')
        # for k,v in import_json(self.sheight.fname_surf).items():
        #     if (k in ['mnab','kx_kz_a_b']):
        #         v = str(v[:10])[:-1] + ', ... (continues) ]'
        #     fprint(f'    {k:15}: {v}')
        # fprint('')
        
        # build grids
        #                                       xyzname,      sheight,      argsP,      argsT,      argsU, fprint)
        self.all_grids        =  All_Grids(case.xyzname, self.sheight, args.argsP, args.argsT, args.argsU, fprint)
        
        # read DNS data
        dreader = Data_Reader( self.all_grids                     ,
                               self.mk_args_reader()              ,
                               fprint                             ,
                               )
        
        # Perform Hex Integration
        fprint( '\n-----------------------------------------------------\n    Staggered Hexahedrons: Quadrature Integration\n-----------------------------------------------------\n')
        Integration_Driver(self, 'full_stag_hex')
        
        # Interpolate to Center (Quadrature Integration)
        fprint( '\n-----------------------------------------------------\n    Interp2Center: Quadrature Integration\n-----------------------------------------------------\n')
        Integration_Driver(self, 'quadint_interp2center')
        
        # Interpolate to Center (Midpoint Integration)
        fprint( '\n-----------------------------------------------------\n    Interp2Center: Midpoint Integration\n-----------------------------------------------------\n')
        Integration_Driver(self, 'midpoint_interp2center')
        
        print(f'Elapsed Time (main): {format_dt(time()-t0)}')
        
        self.info_prof  =  dreader.info_prof
        self.write_main_profiles()
        Plotter_DUTU_Plus( self.mk_args_plotDuTuPlus() )
        print(f'Elapsed Time (Plotter_DUTU_Plus): {format_dt(time()-t0)}')
        
        for side in ('top','bottom'):
            SummaryBalances(self.subfolder_out, side, fprint = fprint, write_tec = True)
        
        print(f'Elapsed Time (all): {format_dt(time()-t0)}')
        
    def ini_fprint(self):
        
        self.logname  =  self.mk_path_ending('.log')
        self.fprint   =  FPrint(self.logname)
    
    def ini_outpath(self):
        
        # initialize time stamp
        self.t_mark  =  tstamp()
        
        # build base prefix (date + tag)
        prefix  =  self.case.tag + '_' + self.t_mark
        
        # create sub-folder path
        self.subfolder_out  =  pjoin(output_dir, prefix)
        
        # assert new subfolder doesn't exist
        assert not isdir(self.subfolder_out)
        
        # create new directory
        mkdir(self.subfolder_out)
        
        # make outpath
        self.outpath         =  lambda x: pjoin(self.subfolder_out, basename(x))
        
        # define function to create paths with a given ending
        self.mk_path_ending  =  lambda end: self.outpath( prefix  +  end )
    
    def write_main_profiles(self):
        info          =  self.info_prof
        profile_path  =  self.mk_path_ending('_profile_statistics.dat')
        header        =  self.mk_header('Main Profile Statistics')
        with open(self.outpath(profile_path),'w') as f:
            write = lambda s: f.write(s+'\n')
            lmap(write, header)
            
            write('\nAverage Bulk Values:')
            for c, value, vol in zip('UVWT', info.bulk_uvwT, info.bulk_vols):
                write(f'    {c}: {qformat(value)}  (vol = {qformat(vol)})')
            
            write('\nVelocity Profile (Streamwise Direction):')
            write('    y                    U(y)')
            for y, value in zip(info.profile_u_y, info.profile_u):
                write(f'    {y:19.12e}  {value:19.12e}')
            write('**End Velocity Profile (Streamwise Direction)**')
            
            write('\nTemperature Profile:')
            write('    y                    T(y)')
            for y, value in zip(info.profile_T_y, info.profile_T):
                write(f'    {y:19.12e}  {value:19.12e}')
            write('**End Temperature Profile**')
            
    def mk_args_reader(self):
        case = self.case
        return as_namedtuple({  'Ret'        :  case.Ret         , 
                                'Pr'         :  case.Pr          ,
                                'tecname'    :  case.tecname     ,
                                'ref_header' :  case.ref_header  ,
                                'ref_zone'   :  case.ref_zone    ,
                                'total_lines':  prod(case.n_xyz) ,
                                'Lx'         :  self.sheight.Lx  ,
                                'subfolder'  :  case.subfolder   ,
                             },'args_reader')
    
    def mk_args_plotDuTuPlus(self):
        return {'data'            : {key: {'profile':  as_list(getattr(self.info_prof,f'profile_{pk}'  ),float),
                                           'y'      :  as_list(getattr(self.info_prof,f'profile_{pk}_y'),float)}
                                                                                                                          for key,pk in zip('UT','uT')},
                'Ly'              :  self.sheight.Ly                                                                                      ,
                'Ret'             :  self.case.Ret                                                                                        ,
                'fig_title'       :  self.case.title                                                                                      ,
                'fig_fname'       :  {key: self.mk_path_ending(f'_D{key}plus.png')                                        for key in 'UT'},
                'dat_fname'       :  {key: self.mk_path_ending(f'_D{key}plus.dat')                                        for key in 'UT'},
                'header'          :  {key: self.mk_header(f'D{key}+ Results for Case: '+self.case.title)                  for key in 'UT'}}
    
    def mk_args_plotSlurm(self, fname_slurm_prefiltered):
        
        return as_namedtuple({ 'fname_slurm_prefiltered'  :  fname_slurm_prefiltered                                         ,
                               't_sweep'                  :  self.info_prof.t_sweep                                          ,
                               'fig_title'                :  self.case.title                                                 ,
                               'fig_fname'                :  self.mk_path_ending('_Slurm_Data.png')                          ,
                               'dat_fname'                :  self.mk_path_ending('_Slurm_Data.dat')                          ,
                               'header'                   :  self.mk_header('Slurm Data Series for Case: '+self.case.title)  ,
                             },'Slurm_args')
    
    def mk_header(self, title):
        
        L_xyz    =  tuple(map( lambda c: getattr(self.sheight,f'L{c}') , 'xyz' ))
        
        nreps_x  =  self.sheight.nreps_x  if hasattr(self.sheight,'nreps_x') else 1
        nreps_z  =  self.sheight.nreps_z  if hasattr(self.sheight,'nreps_z') else 1

        header   =  []
        write    =  lambda s: header.append(s)
        write( '---------------------------------------------------')
        write(f'                {title}')
        write( '---------------------------------------------------')
        write( '\nMain Parameters:')
        write(f'    N_xyz                                  :  {iformat(self.case.n_xyz  )}')
        write(f'    L_xyz                                  :  {qformat(          L_xyz  )}')
        write(f'    Ret                                    :  {qformat(self.case.Ret)}')
        write(f'    Pr                                     :  {qformat(self.case.Pr)}')
        write(f'    nreps_x                                :  {nreps_x}')
        write(f'    nreps_z                                :  {nreps_z}')
        write(f'    Bulk Velocity (Streamwise Direction)   :  {qformat(self.info_prof.bulk_uvwT[ 0])}')
        write(f'    Bulk Temperature                       :  {qformat(self.info_prof.bulk_uvwT[-1])}')
        write(f'    Bulk Flow Through Time (BFTT / t_sweep):  {qformat(self.info_prof.t_sweep      )}')
        
        return header