# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from    collections  import  namedtuple

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from core.utils                    import  ( as_namedtuple ,
                                             qformat       ,
                                             performat     )

from core.single_grid_integration  import  ( Hex_Integration                      ,
                                             Interp2Center_Quadrature_Integration ,
                                             Interp2Center_Midpoint_Integration   )

# ------------------------------------------------------------------------
#                  Basic Functions
# ------------------------------------------------------------------------

def fsum(total, v):
    if type(total)!=list:
        return total+v
    else:
        for i in range(3):
            total[i] += v[i]
        return total

def ferror(total, sheight):
    vol = sheight.Lx * sheight.Ly * sheight.Lz
    if type(total) == list:
        ref  =  [vol, 0., 0.]
        return [ (t-r)/vol*100 for t,r in zip(total,ref) ]
    else:
        return (total/vol-1)*100

# ------------------------------------------------------------------------
#                  Build Load Modes
# ------------------------------------------------------------------------

# The array "Lmode[i,j]" corresponds to a 3x3 matrix indicating which
# functions should be integrated (in the Cartesian direction[i]) along
# the normal component n[j].

# For example, the Pressure force integral:
#     F_p = Int( -P*nx -P*ny -P*nz ) dxdz
#     would have a matrix with diagonal entries:
#         Lmode[i,i] -> {K=-1, mode='f'}
#     This indicates that that the pressure function (standard mode='f')
#     should be integrated along each normal component n[i=j], and
#     multiplied by a constant K = -1.

# The expressions for the heat transfer integral:
#     Q = 1/Ret/Pr * Int( dTdx*nx + dTdy*ny + dTdz*nz ) dxdz
#     also requires diagonal terms for Lmode[i,i] as before.
#     However, the constant "K" now takes a value of (1/Ret/Pr)
#     and the "function mode" now references the index of the normal
#     component n[i] being considered.
#     (nx -> mode=dx, ny -> mode=dy, nz -> mode=dz)

# Finally, the expressions for the viscous Force integral:
#     F_visc = Int( dot(T,n) ) dxdz
#     follows the same rules as before. However, both the integration
#     contants and the "function modes" change following more complex
#     rules. The diagonal entries Lmode[i,i] have a constant K = 2/Ret,
#     whereas non-diagonal entries have a constant K = 1/Ret.
#     The "function modes" employed match the derivatives expected
#     in a Newtonian Stress Tensor.

nt_Lmode   =  namedtuple('Lmode_ij','K,mode,tag_sum,plot,tag_tec')

def complete_tags(Lmode,key):
    for         i,row  in  enumerate(Lmode):
        for     j,obj  in  enumerate(row  ):
            if  obj:
                d             =  obj._asdict()
                d['tag_sum'] +=  f' ({obj.tag_tec}) ({i}) ({key})'
                Lmode[i][j]   =  nt_Lmode(*[d[kk] for kk in obj._fields])

def build_Lmodes( key, Ret, Pr, bool_plot ):
    
    # check key is valid
    assert key in 'UVWPT'
    
    # build Lmode
    Lmode  =  [[None for _ in range(3)] for _ in range(3)]
    
    for      i in  range(3):
        # i: force component
        # j: normal component
        for  j, c in  enumerate('xyz'):
            nj  =  'n' + c
            # case P
            if (key == 'P') and (i==j):
                # int ( -P * n ) dx dz
                Lmode[i][j] = nt_Lmode( K        = -1.              ,
                                        mode     = 'f'              ,
                                        tag_sum  =  f'-P*{nj}'      ,
                                        plot     =  bool_plot       ,
                                        tag_tec  = f'minus_P_{nj}'  )
            # case T
            if (key == 'T') and (i==j):
                # int ( 1/Ret/Pr * grad_T * n ) dx dz
                Lmode[i][j] = nt_Lmode( K        = 1./(Ret*Pr)            ,
                                        mode     = f'd{c}'                ,
                                        tag_sum  =  f'1/Pet*dTd{c}*{nj}'  ,
                                        plot     =  bool_plot             ,
                                        tag_tec  = f'inv_Pet_dTd{c}_{nj}' )
            
            # case U,V,W
            if (key in 'UVW'):
                
                # int( 1/Ret dot(tau,n) )
                
                vel_ind = 'UVW'.index(key)
                
                #  check if row or column match the velocity component
                # (otherwise, it can be proven that Lmode[i,j] = null)
                
                if vel_ind in [i,j]:
                    # add multiplier of 2 for diagonal components
                    mag = 2. if (i==j) else 1.
                    
                    # when vel==i, we differentiate along the current
                    # j component, otherwise, we consider the "i" direction.
                    # (This formula can be easily verified by analyzing all
                    # components of a Newtonian Stress Tensor.)
                    
                    d   = 'xyz'[j if (vel_ind==i) else i]
                    Lmode[i][j] = nt_Lmode( K       =  mag/Ret                          ,
                                            mode    =  f'd{d}'                          ,
                                            tag_sum =  f'{mag}/Ret*d{key}d{d}*{nj}'     ,
                                            plot    =  bool_plot                        ,
                                            tag_tec =  f'inv_Ret_{mag}_d{key}d{d}_{nj}' )
    #
    complete_tags(Lmode,key)
    return Lmode

# ------------------------------------------------------------------------
#                  Main Routine
# ------------------------------------------------------------------------

# Integration_Driver: prints a detailed report of the force and 
#     heat transfer balances obtained after integrating all surface locations
#     according to a framework (full_hex,interp2center).
#     The orders to the underlying "framework" are dispatched by internal commands,
#     which only receive the summed force or heat transfer balance obtained
#     during each call.

class Integration_Driver:
    def __init__(self, par, mode):
        
        # par is the parent object (from "engine.py")
        
        # register input
        self.par, self.mode = par, mode
        
        # fetch integration function
        
        self.is_interp2center = 'interp2center' in mode
        
        if self.is_interp2center:
            grid_integral_function  =  {  'quadint_interp2center' : Interp2Center_Quadrature_Integration  ,
                                          'midpoint_interp2center': Interp2Center_Midpoint_Integration    }[mode]
        else:
            grid_integral_function  =  {  'full_stag_hex'         : Hex_Integration                       }[mode]
        
        # build final dictionary
        Final = { 'top': {} , 'bottom': {} }
        
        # integrate by variable, registering the results
        for side in ('top','bottom'):
            for key in 'UVWPT':
                
                # fetch mini-report (Dsum)
                Dsum     =  grid_integral_function(  par.all_grids(key)           ,       #  fetch grid
                                                     self.mk_args_int(side, key)  ).Dsum  #  run-arguments

                # print new results, and register in Final
                self.detailed_report(Dsum, side, key, Final)
        
        # print final balance
        self.final_balance(Final)
        
    def detailed_report(self, Dsum, side, key, Final):
        
        # total: computes the net balance in the xyz directions
        total   =  [0,0,0]
        
        # print type of magnitude (force/heat transfer)
        assert key in 'UVWPT'
        nature  =  "Forces" if key != "T" else "Heat Transfer"
        self.par.fprint(f'\nDetailed {nature} (mode = {self.mode}) (side = {side}) (type = {key})\n')
        
        # print all tagged components:
        for i,D in enumerate(Dsum):
            for k,v in D.items():
                self.par.fprint(f'    [{i}] {k}: {v:16.5f}')
                total[i] += v
        
        # Final: register total force balance per side
        #        e.g. Final['bottom']['U'] = [3.14, 0.1, 0.03]
        
        # store total by component
        if key in 'UVWP':
            # assert that key has not been processed before
            assert not key in Final[side]
            Final[side][key] = total
        else:
            assert key == 'T'
            # for temperature, store values separated by direction (xyz)
            for i,c in enumerate('xyz'):
                k = f'dTd{c}'
                # again, assert that key did not exist before
                assert not k in Final[side]
                Final[side][k] = total[i]
    
    def final_balance(self, Final):
        
        fprint = self.par.fprint
        fprint('\n----------------------------------------------------')
        fprint( f'      Final Summary (mode = {self.mode})')
        fprint(  '----------------------------------------------------')
        
        for is_force in (True,False):
            # inform whether forces or heat transfer is being printed (this cycle)
            title_    =  'Forces' if is_force else 'Heat Transfer'
            fprint(f'\n    {title_}:')
            
            # consider "global" as a "third side"
            for side in ('top','bottom','global'):
                
                fprint(f'\n        {side.title()}:')
                
                # track total by side
                total = [0,0,0] if is_force else 0
                
                # make relevant keys
                all_keys = 'UVWP' if is_force else list(map(lambda c: f'dTd{c}','xyz'))
                
                # loop over keys, printing values
                for key in all_keys:
                    
                    # compute value for [side][key] (summing if side=global)
                    if side != 'global':
                        values = Final[side][key]
                    
                    else:
                        # for global, sum over both sides
                        values = [0,0,0] if is_force else 0
                        for s in ('top','bottom'):
                            values = fsum(values, Final[s][key])
                    
                    # print value, by component, by side
                    fprint(f'                {key}: {qformat(values)}')
                    # track global total
                    total  =  fsum(total,values)
                fprint(    f'\n            Total: {qformat(total)}')
                # print error comparison, if side == global
                if side == 'global':
                    fprint(  f'            Error: {performat(ferror(total, self.par.sheight))}')
        fprint('')
        assert ('top' in Final) and ('bottom' in Final)
    
    def mk_args_int(self, side, key, bool_plot = True):
        
        par      =  self.par
        Ret, Pr  =  par.case.Ret, par.case.Pr
        
        D = {  'Ret'        :  Ret                                                   , 
               'Pr'         :  Pr                                                    ,
               'Lx'         :  par.sheight.Lx                                        ,
               'Lz'         :  par.sheight.Lz                                        ,
               'key'        :  key                                                   ,
               'side'       :  side                                                  ,
               'tec2dname'  :  par.mk_path_ending(f'_{self.mode}_{side}_{key}.tec')  ,
               'Lmode'      :  build_Lmodes( key, Ret, Pr, bool_plot )               ,
            }
        
        get_Pxz_attr  =  lambda sufix: [ getattr(par.all_grids('P'), f'{c}{sufix}') for c in 'xz' ]
        D['xp_2d_centers'], D['zp_2d_centers']  =  get_Pxz_attr('_2d')
        D['xp_1d_centers'], D['zp_1d_centers']  =  get_Pxz_attr('1')
        D['yp_2d_centers']                      =  getattr(par.all_grids('P'), f'y_{side}')
        D['sheight']  =  par.sheight
            
        return as_namedtuple(D,'args_integration')