# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------
import  torch
from    os.path  import  join  as  pjoin
from    os.path  import  basename
# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from core.io_access  import  data_dirs
from core.utils      import  reader, lmap, lfilter, as_list

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

zeros_like    =  torch.zeros_like
searchsorted  =  torch.searchsorted

lrange        =  lambda a,b,n: list(range(a,b,n))
spacesplit    =  lambda x    : lfilter(None,x.split(' '))

fmin          =  lambda a,b  : a  if a<b else  b
fmax          =  lambda a,b  : a  if a>b else  b

# searchsorted(a, v, right = ?)
#   left:             v =< a[i]
#   right:  a[i-1] <= v

find_first = lambda a,v: searchsorted(a, v, right=False).item()
find_last  = lambda a,v: searchsorted(a, v, right=True ).item() - 1

# get_index: find index A[i], where array matches "c".
#            Assert that only one match was found.
def get_index(A, tag):
    j = [i for i,x in enumerate(A) if (tag in x)]
    assert len(j) == 1
    return j[0]

# tolong: convert list of integers into torch.tensor
tolong  =  lambda x: torch.tensor(x, dtype = torch.long)

# pfix: find neighboring elements in an array, accounting for
#       periodicity (0+x-> L+x)
def pfix(arr, ind, L):
    size = len(arr)
    if ind<0:
        # return last element, but shifted (back) by L
        return arr[ind+size]-L
    elif ind>=size:
        # return first element, but shifted (fwd) by L
        return arr[ind-size] + L
    else:
        return arr[ind]

# ------------------------------------------------------------------------
#                  Import XYZ data
# ------------------------------------------------------------------------

class XYZ_data:
    def __init__(self, xyzname):
        
        self.__coords = {}
        
        # read input file with "xyz"
        A  =  reader(xyzname)
        
        # extract xp, xu, etc.
        self.extract(A,'xp_xu[k]')
        self.extract(A,'yp_yu[i]')
        self.extract(A,'zp_zu[j]')
       
    def __call__(self, c):
        return self.__coords[c]
    
    def extract(self, A, tag):
        # get indexes start/end
        i1 = get_index( A , 'BEGIN_' + tag )
        i2 = get_index( A , 'END_'   + tag )
        
        # split data by space
        p,u = self.tensor_cols(A[i1+1:i2])
        
        # assert shapes
        assert len(p.shape) == len(u.shape) == 1
        assert     p.shape  ==     u.shape
        
        assert (list(p) == sorted(p)) and (list(u) == sorted(u))
        
        assert all( torch.diff(arr).min().item()>1e-10 for arr in [p,u] )
        
        # automatically create new attributes
        tp,tu = (tag.split('[')[0]).split('_')
        
        self.__coords[tp] = p
        self.__coords[tu] = u
    
    def tensor_cols(self, lines):
        f     =  lambda x: torch.tensor(x, dtype = torch.double)
        A, B  =  [], []
        for x in lines:
            a,b = lmap(float,spacesplit(x))
            A += [a]
            B += [b]
        return f(A),f(B)

# ------------------------------------------------------------------------
#                  Grid 2-D
# ------------------------------------------------------------------------

def grid2d(x1,z1):
    shape  =  len(x1), len(z1)
    X      =  torch.zeros(shape, dtype = torch.double)
    Z      =  torch.zeros(shape, dtype = torch.double)
    
    for k,x_  in  enumerate(x1):
        X[k,:] = x_
        
    for j,z_  in  enumerate(z1):
        Z[:,j] = z_
    
    return X,Z

# ------------------------------------------------------------------------
#                  Grid 3-D
# ------------------------------------------------------------------------

class Grid3d:
    def __init__(self, x1, y1, z1, seq, sheight, args, tag, fprint):
        
        self.x1    =  x1
        self.y1    =  y1
        self.z1    =  z1
        self.seq   =  seq
        self.args  =  args
        self.tag   =  tag
        n_side     =  args.n_side
        
        shape2        =  self.shape2  =  len(x1),len(z1)
        
        self.i_bottom                 =  torch.zeros( (*shape2, n_side  ), dtype = torch.long    )
        self.i_top                    =  torch.zeros( (*shape2, n_side  ), dtype = torch.long    )
        self.vals_bottom              =  torch.zeros( (*shape2, n_side  ), dtype = torch.double  ) * float('nan')
        self.vals_top                 =  torch.zeros( (*shape2, n_side  ), dtype = torch.double  ) * float('nan')
        
        self.find_y_valid(sheight, fprint)
        
        self.mk_get_vol(sheight)
        
    def find_y_valid(self, sheight, fprint):
        
        self.x_2d, self.z_2d  =  grid2d(self.x1, self.z1)
        
        shape2  =  self.x_2d.shape
        assert  self.x_2d.shape  ==  self.z_2d.shape  ==  shape2
        
        self.y_bottom  =  y_bottom  =  sheight.get_h(self.x_2d, self.z_2d, side='bottom', lib=torch)
        self.y_top     =  y_top     =  sheight.Ly - y_bottom
        
        y1, tol, n_side  =  self.y1, self.args.tol, self.args.n_side
        tol                       =  eval(tol)(torch.diff(self.y1).min().item())
        assert tol > 1e-10
        
        fprint(f'\nGrid Type: {self.tag}')
        fprint(f'    Tolerance Function : {self.args.tol}')
        fprint(f'    Tolerance Evaluated: {tol:12.6f}')
        
        i_bottom        =  self.i_bottom      
        i_top           =  self.i_top         
         
        
        # indexing order is (ijk -> yzx) (xyz -> kij)
        
        def mk_range( val, n, find, k = 1, extremes = False):
            assert (k in (1,-1)) and n>0
            i  =  find(y1, val)
            if extremes:
                return  tolong(sorted([ i , i+(n-1)*k ]))
            else:
                return  tolong(lrange(  i , i + n*k   , k ))
        
        for     k  in  range(shape2[0]):
            for j  in  range(shape2[1]):
                
                i_bottom      [k,j,:]  =  mk_range( y_bottom[k,j] + tol , n_side  , find_first     )
                i_top         [k,j,:]  =  mk_range( y_top   [k,j] - tol , n_side  , find_last , -1 )
    
    def mk_vol_dx(self, sheight):
        Lx  =  sheight.Lx
        x1  =  self.x1
        
        self.vol_dx  =  zeros_like( x1 )
        for k in range(len(x1)):
            #            (      x_right    -      x_left     ) / 2
            self.vol_dx[k] = ( pfix(x1,k+1,Lx) - pfix(x1,k-1,Lx) ) * 0.5
    
    def mk_vol_dz(self, sheight):
        Lz           =  sheight.Lz
        z1           =  self.z1
        
        self.vol_dz  =  zeros_like( z1 )
        for j in range(len(z1)):
            #            (      z_right    -      z_left     ) / 2
            self.vol_dz[j] = ( pfix(z1,j+1,Lz) - pfix(z1,j-1,Lz) ) * 0.5
    
    def mk_vol_dy(self):
        
        extrapolate  =  lambda a,b: [2*a-b]
        y_edges      =  as_list(self.y1[:-1] + 0.5*torch.diff(self.y1), float)
        y_edges      =  torch.tensor((extrapolate(y_edges[ 0],y_edges[ 1])  +
                                      y_edges                               +
                                      extrapolate(y_edges[-1],y_edges[-2])  ),
                                      dtype = torch.double)
        
        self.y_corner_down  =  y_edges[ :-1]
        self.y_corner_up    =  y_edges[1:  ]
    
    def mk_get_vol(self, sheight):
        
        self.mk_vol_dx(sheight)
        self.mk_vol_dz(sheight)
        self.mk_vol_dy()
        
        vol_dx         =  as_list(self.vol_dx        , float)
        vol_dz         =  as_list(self.vol_dz        , float)
        y_top          =  as_list(self.y_top         , float)
        y_bottom       =  as_list(self.y_bottom      , float)
        y_corner_up    =  as_list(self.y_corner_up   , float)
        y_corner_down  =  as_list(self.y_corner_down , float)
        
        dxz            =  as_list(zeros_like(self.y_top) , float)
        
        for     k,vdx  in  enumerate(vol_dx):
            for j,vdz  in  enumerate(vol_dz):
                dxz[k][j] = vdx*vdz
        
        def get_vol(k,i,j):
            # prevent upper limit higher than y_top
            y2   = fmin(y_top   [k][j], y_corner_up  [i] )
            # prevent lower limit below y_bottom
            y1   = fmax(y_bottom[k][j], y_corner_down[i] )
            # return volume, only if (y2-y1) is positive
            return dxz[k][j]*fmax(0,y2-y1)
        
        self.get_vol = get_vol

# ------------------------------------------------------------------------
#                  Make All Grids
# ------------------------------------------------------------------------

class All_Grids:
    def __init__(self, xyzname, sheight, argsP, argsT, argsU, fprint):
        
        self.xyz  =  XYZ_data(xyzname)
        fetch     =  lambda seq: [self.xyz(c+s) for c,s in zip('xyz',seq)] + [seq]
        
        #  Main Format:
        #     P/T[i,j,k]  ->  xp[k] yp[i] zp[j]
        #       u[i,j,k]  ->  xu[k] yp[i] zp[j] (W in fortran)
        #       v[i,j,k]  ->  xp[k] yu[i] zp[j] (U in fortran)
        #       w[i,j,k]  ->  xp[k] yp[i] zu[j] (V in fortran)
        
        fprint('\n--------------------------------------------------\n    Building Grids\n--------------------------------------------------\n')
        
        self.__dgrids  =  D  =  {}
        
        D['P']   =  Grid3d(*fetch('ppp'), sheight, argsP,'P',fprint)
        D['T']   =  Grid3d(*fetch('ppp'), sheight, argsT,'T',fprint)
        
        # note change uvw -> WUV
        D['U']   =  Grid3d(*fetch('upp'), sheight, argsU,'U',fprint)
        D['V']   =  Grid3d(*fetch('pup'), sheight, argsU,'V',fprint)
        D['W']   =  Grid3d(*fetch('ppu'), sheight, argsU,'W',fprint)
        
    def __call__(self, c):
        return self.__dgrids[c]
