# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import torch

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

zeros_like = torch.zeros_like

# ------------------------------------------------------------------------
#                  Shape Function Class
# ------------------------------------------------------------------------

class ShapeFunction:
    def __init__(self, all_X, all_Y, all_Z, all_U, nx, ny, nz):
        
        # this class will fit a shape function of order (nx,ny,nz)
        # to every batch of of coordinates contained in all_?
        #
        # as a reference, a quadratic shape function is given by n=2
        # the stacks of points all_? contain a list of slices 
        # all_? = [ ?2d_a, ?2d_b, ...] such that they correspond to the
        # points required to fit a shape function at multiple 2d locations
        # at the same time
        # thus, len(all_?) is always (nx+1)(ny+1)(nz+1)
        
        self.nx, self.ny, self.nz = nx, ny, nz
        
        shape2 =  all_U[0].shape
        m_xyz  =  (nx+1)*(ny+1)*(nz+1)
        
        # assert shape was 2D
        assert len(shape2) == 2
        
        # run a quick input check (without leaving intermediate variables)
        def check_input():
            for all_v in [all_X, all_Y, all_Z, all_U]:
                # check total number of elements
                assert len(all_v) == m_xyz
                # find unique shape(s)
                assert all( a.shape == shape2 for a in all_v )
        check_input()
        
        # find avg values
        
        self.__set_avg(all_X, 'X')
        self.__set_avg(all_Y, 'Y')
        self.__set_avg(all_Z, 'Z')
        
        # find constants
        # matrix "A" is a 4-D tensor with shape [:,:,m,m]
        # vector "b" is a 3-D tensor with shape [:,:,m]
        # the shape function terms are grouped as:
        # bi = b[:,:,i] = A[:,:,i,:] * constants[:,:,:]
        # thus, for every bi element, the terms[i,:] are indexed by the last dimension
        # this is consistent with the notion of recovering a normal system of equations
        # at a given [x,y] 2-D location
        
        A = torch.zeros( (*shape2, m_xyz, m_xyz), dtype = torch.double)
        b = torch.zeros( (*shape2, m_xyz)       , dtype = torch.double)
        
        for  i in range(m_xyz):
            b[:,:,i]   =  all_U[i]
            for j, tij in enumerate(self.__get_terms(  all_X[i]  ,
                                                       all_Y[i]  ,
                                                       all_Z[i]  ,
                                                       'f'       )):
                A[:,:,i,j] =  tij
        # print('A',str(as_list(A,float)))
        # print('b',str(as_list(b,float)))
        constants       =  torch.linalg.solve(A,b)
        self.constants  =  [ constants[:,:,j] for j in range(m_xyz) ]
    
    def __get_terms(self, X_, Y_, Z_, mode):
        
        X = X_ - self.X_avg
        Y = Y_ - self.Y_avg
        Z = Z_ - self.Z_avg
        
        get_ai = lambda i: i
        get_aj = lambda j: j
        get_ak = lambda k: k
        
        if    mode == 'f' :
            get_K = lambda i,j,k:   1.
        
        elif  mode == 'dx':
            assert self.nx >= 1
            get_K  = lambda i,j,k: i
            get_ai = lambda i    : i-1
        
        elif  mode == 'dy':
            assert self.ny >= 1
            get_K  = lambda i,j,k: j
            get_aj = lambda   j  : j-1
        
        elif  mode == 'dz':
            assert self.nz >= 1
            get_K  = lambda i,j,k: k
            get_ak = lambda     k: k-1
        
        else:
            raise Exception(f'Unrecognized (mode = {mode}')
        
        terms = []
        for          i in  range(self.nx + 1):
            for      j in  range(self.ny + 1):
                for  k in  range(self.nz + 1):
                    K  = get_K(i,j,k)
                    if K:
                        terms.append( K*(X**get_ai(i)) *
                                        (Y**get_aj(j)) * 
                                        (Z**get_ak(k))  )
                    else:
                        terms.append( zeros_like(X) )
        return terms
    
    def __call__(self, X, Y, Z, mode):
        return sum( c*t for c,t in zip( self.constants                  ,
                                        self.__get_terms(X, Y, Z, mode) ))
    
    def __set_avg(self, all_v, tag):
        setattr(  self                   ,
                  f'{tag}_avg'           ,
                  sum(all_v)/len(all_v)  )