# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  torch
from    math         import  ceil

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.utils           import  lmap, as_list
from  core.misc_tools      import  Gauss_Legendre_Integration, Write_Tec2d
from  core.shape_function  import  ShapeFunction

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

ones_like    =  torch.ones_like

def padded_ends(c1):
    
    c1           =  as_list(c1, float)
    extrapolate  =  lambda a,b: [2*a-b]
    
    # make c1_padded
    c1_padded    =  (  extrapolate(c1[ 0], c1[ 1])  +
                       c1                           +
                       extrapolate(c1[-1], c1[-2])    )
    
    assert c1_padded == sorted(c1_padded)
    
    return  torch.tensor(c1_padded, dtype = torch.double)

def grad_h(x, z, get_h, adim_tol = 5e-3):
    assert x.dtype == z.dtype == torch.double
    h        =   get_h(x,z)
    get_tol  = lambda arr: max([float(torch.mean(torch.diff(arr, dim=k))) for k in range(len(arr.shape))]) * adim_tol
    tol_x    =  get_tol(x)
    tol_z    =  get_tol(z)
    grad_x   =  (get_h(x+tol_x, z      ) - get_h(x-tol_x, z      ))/(2*tol_x)
    grad_z   =  (get_h(x      , z+tol_z) - get_h(x      , z-tol_z))/(2*tol_z)
    return h, grad_x, grad_z

# ------------------------------------------------------------------------
#                  Roll Function
# ------------------------------------------------------------------------

def roll_1d(x,k,L):
    x = torch.roll(torch.clone(x),k)
    if k<0:
        # array (x) was pulled "towards the origin"
        # therefore, the last x[-|k|:,:] positions must be
        # shifted by L0
        x[k:] += L
    else:
        # opposite scenario:
        # array (x) was pulled "towards the end"
        # the fist "k" elements must be substracted L0
        x[:k] -= L
    return x

def roll(x, k, a, L0,L1):
    
    # assert that array "x" is 2d
    assert len(x.shape) == 2
    
    # assert L0 and L1
    if L0:
        assert (L1 is None) and (a==0)
    if L1:
        assert (L0 is None) and (a==1)
    
    if k:
        x = torch.roll(torch.clone(x), k, dims = a)
        
        if   (a == 0) and L0:
            # check that array "x" varies along each row
            assert x[0,0] != x[1,0] and x[0,0] == x[0,1]
            
            if k<0:
                # array (x) was pulled "towards the origin"
                # therefore, the last x[-|k|:,:] positions must be
                # shifted by L0
                x[k:,:] += L0
            else:
                # opposite scenario:
                # array (x) was pulled "towards the end"
                # the fist "k" elements must be substracted L0
                x[:k,:] -= L0
        
        elif (a == 1) and L1:
            
            # check that array "x" varies along each column
            assert x[0,0] == x[1,0] and x[0,0] != x[0,1]
            
            # similar cases as before, but for the inner dimension
            if k<0:
                x[:,k:] += L1
            else:
                x[:,:k] -= L1
    return x

# ------------------------------------------------------------------------
#                  Roll Indexes
# ------------------------------------------------------------------------

def find_roll_indexes(cen, c1):
    
    assert cen.shape == c1.shape

    c1_padded  =  padded_ends(padded_ends(c1))
    assert (c1_padded[0]<cen[0]) and (cen[-1]<c1_padded[-1])
    
    # find lower, higher
    #     sorted_sequence    right
    #     1-D                False   A[i-1] < values
    #     1-D                True             values < A[i]
    
    # find first index in c1_padded lower/higher than cen[?]
    lower      =  torch.searchsorted(c1_padded, cen-1e-10, right = False)-1 # last  index in c1_padded that is lower  than "cen" (i-1)
    higher     =  torch.searchsorted(c1_padded, cen+1e-10, right = True )   # first index in c1_padded that is higher than "cen" (i)
    
    # substract one, to account for extra element given by [c1_ini]
    lower     -=  2
    higher    -=  2
    
    # find relative offset between index-position in "cen" and found "c1" positions in "lower/higher"
    equal      =  torch.tensor( range(len(cen)) , dtype = torch.int64 )
    
    lower      =  torch.unique( lower  - equal )
    higher     =  torch.unique( higher - equal )
    
    assert lower.numel() == higher.numel() == 1
    
    lower      =  lower.item()
    higher     =  higher.item()
    
    assert            lower < higher
    
    # remember that a higher index must be fetched by a "negative roll" (pull towards origin)
    # therefore, the new values of [lower,higher] must be reversed to obtain all_k_roll
    return [-val for val in range(lower, higher+1)]

# ------------------------------------------------------------------------
#                  Call XZ_amp and DXDZ
# ------------------------------------------------------------------------

def get_XZamp_dxdz(x_cen, z_cen, Lx, Lz):
        
        # compute x_amp, z_amp (distance from center to one extreme of the element)
        
        # x_prev      -      x_cen      -      x_next
        #             | x_amp  -  x_amp |
        # x_amp is 25% of the total distance
        
        x_amp    =  0.25*( roll(x_cen, -1, 0, Lx  , None) -   # 0.25*( x_next -
                           roll(x_cen, +1, 0, Lx  , None)   ) #        x_prev   )
        
        z_amp    =  0.25*( roll(z_cen, -1, 1, None, Lz  ) -   # 0.25*( z_next -
                           roll(z_cen, +1, 1, None, Lz  )   ) #        z_prev   )
        
        assert (x_amp>0).all() and (z_amp>0).all()
        
        #    dxdz_2d  =  (2*x_amp) * (2*z_amp)
        dxdz_2d  =   4 * x_amp * z_amp
        
        return x_amp, z_amp, dxdz_2d

# ------------------------------------------------------------------------
#                  Base Integration Class
# ------------------------------------------------------------------------

class Base_Grid_Integration_Class:
    def __init__(self, grid, args):
        
        self.grid  =  grid
        self.args  =  args
        
        # detect side
        side  =  args.side
        assert side in ('top','bottom')
        self.is_bottom = (side == 'bottom')
        
        # note: "split_parts=false" may be overwritten by hex_integration
        self.split_parts = False
        
    def run(self):
        
        args          =  self.args
        get_h         =  self.args.sheight.get_h
        
        # fetch Lmode
        Lmode         =  args.Lmode
        
        all_tec_tags  =  [ 'X'       ,  'Y'       , 'Z'        , 'H'                                                   ]
        all_tec_2d    =  [ self.x_cen,  self.y_cen, self.z_cen , get_h(self.x_cen, self.z_cen, side='bottom',lib=torch)]
        
        if self.split_parts:
            corners_tec_tags  =  [ 'X'           ,  'Y'           , 'Z'            , 'H'                                                           ]
            corners_tec_2d    =  [ self.x_corners,  self.y_corners, self.z_corners , get_h(self.x_corners, self.z_corners, side='bottom',lib=torch)]
        
        self.Dsum  =  Dsum  =  [{},{},{}]
        
        for i in range(3):
            for j in range(3):
                if Lmode[i][j]:
                    lm       =  Lmode[i][j]
                    
                    # f(x,z)  =  nj(x,z) * K * b_fit(x, h(x,z), z , mode)
                    
                    if not self.split_parts:
                        values                 =  self.get_values(j, lm.K, lm.mode)
                        values_sum             =  values.sum().item()
                    else:
                    
                        values, values_corners =  self.get_values(j, lm.K, lm.mode)
                        values_sum             =  values.sum().item()
                        
                        assert abs( values_sum - values_corners.sum().item() )<1e-8, (values_sum, values_corners.sum().item())
                        
                    
                    assert not lm.tag_sum in Dsum[i]
                    Dsum[i][lm.tag_sum]  =  values_sum
                    
                    if lm.plot:
                        all_tec_tags.append(lm.tag_tec)
                        all_tec_2d  .append(values)
                        
                        if self.split_parts:
                            corners_tec_tags.append(lm.tag_tec)
                            corners_tec_2d  .append(values_corners)
        
        if len(all_tec_2d) > 4:
            
            all_tec_tags  +=  [ 'dxdz'       ]
            all_tec_2d    +=  [ self.dxdz_2d ]
            
            Write_Tec2d( all_tec_tags   ,
                         all_tec_2d     ,
                         args.tec2dname )
            
            if self.split_parts:
                
                corners_tec_tags  +=  [ 'dxdz'               ]
                corners_tec_2d    +=  [ self.dxdz_2d_corners ]
                
                Write_Tec2d( corners_tec_tags                              ,
                             corners_tec_2d                                ,
                             args.tec2dname.replace('.tec','_corners.tec') )
    
    def make_b_fit(self):
        grid       =  self.grid
        gather_2d  =  self.gather_2d
        
        # interior layer of values (like i_{side}, vals_{side})
        from_side  =  lambda name: getattr(grid, f'{name}_{self.args.side}')
        
        i_side     =  from_side('i'   )
        vals_side  =  from_side('vals')
        
        layers     =  vals_side.shape[-1]
        
        # find 2d-plane degree
        nx         = self.fit_nx
        nz         = self.fit_nz
        
        assert nx == (len(self.all_kx_roll)-1)
        assert nz == (len(self.all_kz_roll)-1)
        
        # build arrays of points
        all_Y   = []
        all_U   = []
        
        # check whether to add walls, or only interior points
        assert grid.tag in 'UVWPT'
        
        if grid.tag == 'P':
            assert layers == 2
            ny = 1
        else:
            # wall velocities
            all_Y   +=  gather_2d( from_side('y') )
            all_U   +=  lmap(torch.zeros_like, all_Y)
            ny       =  layers
        
        assert all( (len(arr.shape)==3 and (arr.shape[-1] == layers)) for arr in [i_side,vals_side] )
        
        # base 2d grid
        all_X  =  gather_2d( grid.x_2d , L0 = self.args.Lx) * (ny+1)
        all_Z  =  gather_2d( grid.z_2d , L1 = self.args.Lz) * (ny+1)
        
        # add interior values (pressure are added here too)
        for pos in range(layers):
            all_Y  +=  gather_2d( grid.y1[ i_side   [:,:,pos] ] )
            all_U  +=  gather_2d(          vals_side[:,:,pos]   )
        
        return ShapeFunction(all_X, all_Y, all_Z, all_U, nx, ny, nz)
    
    def gather_2d(self, p, L0 = None, L1 = None):
        
        assert len(p.shape) == 2
        
        result  = []
        for      k0 in  self.all_kx_roll:
            x = roll(p, k0, 0, L0,None)
            for  k1 in  self.all_kz_roll:
                result.append(roll(x, k1, 1 ,None,L1))
        return result
    
    def assert_almost(self, fit):
        for c in 'xz':
            a        =  getattr(self, f'{c}_cen')
            b        =  getattr(fit , f'{c.upper()}_avg')
            residual =  torch.abs(a-b).max().item() 
            if residual >= 1e-10:
                print(f'Warning: (residual = {residual})')
        assert residual < 1e-6

# ------------------------------------------------------------------------
#                  Gauss-Legendre Quadrature Integration
# ------------------------------------------------------------------------

class Quadrature_Integration(Base_Grid_Integration_Class):
    def __init__(self, grid, args, extract_xz_cen = False, split_parts = False):
        
        super().__init__(grid, args)
        
        # build fits:
        h_fit      =  self.make_h_fit()
        b_fit      =  self.make_b_fit()
        
        # wrapper height function: (y-coordinate is cancelled out)
        h          =  lambda x, z, mode:  h_fit(x, 1., z, mode)
        
        # compute normal:
        
        #     flipn      =  1 if self.is_bottom else -1
        
        #     at the bottom side, the normal has the right orientation
        #     however, at the top, the normal is calculated with mirrored points,
        #     and it points inwards. therefore, the ENTIRE vector must be reversed.
        
        #     note: when we calculated the normal through the analytical function,
        #           only the n[1] component was reversed (unlike here), because
        #           the reference calculation  points were also at the bottom surface
        
        if self.is_bottom:
            
            n = [  lambda x,z:  -h(x, z, 'dx')  ,
                   lambda x,z:   1.0            ,
                   lambda x,z:  -h(x, z, 'dz')  ]
        
        else:
            # reversed formula, top side
            n = [  lambda x,z:   h(x, z, 'dx')  ,
                   lambda x,z:  -1.0            ,
                   lambda x,z:   h(x, z, 'dz')  ]
        
        # cell centers
        if extract_xz_cen:
            x_cen  =  self.x_cen  =  h_fit.X_avg
            z_cen  =  self.z_cen  =  h_fit.Z_avg
        else:
            x_cen  =  self.x_cen
            z_cen  =  self.z_cen
           
        self.x_amp, self.z_amp, self.dxdz_2d = get_XZamp_dxdz(x_cen, z_cen, args.Lx, args.Lz)
        
        # gauss-legendre integration points
        ox  =  ceil((h_fit.nx + b_fit.nx + 1)*0.5)
        oz  =  ceil((h_fit.nz + b_fit.nz + 1)*0.5)
        
        assert ox == ceil(( 2*self.fit_nx + 1 )*0.5)
        assert oz == ceil(( 2*self.fit_nz + 1 )*0.5)
        
        self.assert_almost(h_fit)
        self.assert_almost(b_fit)
        
        self.split_parts  =  split_parts

        self.gl_scheme    =  Gauss_Legendre_Integration( x_cen , self.x_amp , ox ,
                                                         z_cen , self.z_amp , oz )
        
        if self.split_parts:
            self._ini_corners()
        
        # function to "get values" in main runner
        def get_values(j, K, mode):
            
            nj     =  n[j]
            # f(x,z) =              nj(x,z) * K * b    (x, h(x,z)     ,z, mode)
            f        =  lambda x,z: nj(x,z) * K * b_fit(x, h(x,z,'f'), z, mode)
            
            if self.split_parts:
                
                parts = self.gl_scheme.integral_parts(f)
                
                return ( sum( parts )                  ,
                         self.sum_parts2corners(parts) )
            
            else:
                return  self.gl_scheme.integrate_f(f)
        
        self.get_values = get_values
        
        # register cell-centers for later usage
        # self.y_cen = h(x_cen, z_cen, 'f')
        if self.is_bottom:
            self.y_cen   =                         self.args.sheight.get_h(x_cen, z_cen, side='bottom', lib=torch)
        else:
            self.y_cen   =  self.args.sheight.Ly - self.args.sheight.get_h(x_cen, z_cen, side='bottom', lib=torch)
        
        self.run()
    
    def make_h_fit(self):
        
        grid  =  self.grid
        
        # bilinear element (x,z):
        nx = self.fit_nx
        nz = self.fit_nz
        
        # verify that (nx=nz=1) matches all_k_roll
        assert (len(self.all_kx_roll)-1) == nx
        assert (len(self.all_kz_roll)-1) == nz
        
        # y direction is constant:
        ny      = 0
        
        all_X   = self.gather_2d( grid.x_2d , L0 = self.args.Lx)
        all_Y   = lmap( ones_like, all_X )
        all_Z   = self.gather_2d( grid.z_2d , L1 = self.args.Lz)
        all_U   = self.gather_2d( grid.y_bottom if  self.is_bottom  else grid.y_top )
        
        return ShapeFunction(all_X, all_Y, all_Z, all_U, nx, ny, nz)

# ------------------------------------------------------------------------
#                  Hex Integration (Staggered Coords.)
# ------------------------------------------------------------------------

class Hex_Integration(Quadrature_Integration):
    def __init__(self, grid, args):
        self.all_kx_roll  =  self.all_kz_roll  =  [0, -1]
        self.fit_nx       =  self.fit_nz       = 1 
        
        super().__init__(grid, args, extract_xz_cen = True, split_parts = True)
    
    def _ini_corners(self):
        
        self.x_corners               =  self.args.xp_2d_centers
        self.y_corners               =  self.args.yp_2d_centers
        self.z_corners               =  self.args.zp_2d_centers
        
        _, _, self.dxdz_2d_corners   =  get_XZamp_dxdz( self.x_corners , self.z_corners ,
                                                        self.args.Lx   , self.args.Lz   )
        
        self.all_rolls_parts         =  [ self.__find_rolls_parts(xi,zi) for xi,zi in self.gl_scheme.get_coords_xz() ]
        
    def sum_parts2corners(self, parts):
        # roll "parts" to sum them to the right positions in "corners"
        s = 0.
        for ch, all_rr in zip(parts, self.all_rolls_parts):
            rx, rz = all_rr
            s     += roll( roll(ch,rx,0,None,None) ,
                                   rz,1,None,None  ) # L0=L1=None because we are rolling data
        return s
    
    def __find_rolls_parts(self, xi, zi):
        # find position in xz_corners that is closest to gauss_legendre_points (xi,zi)

        # check XZ arrangement
        check_X = lambda x: len(x.shape)==2 and (x[0,0]!=x[1,0]) and (x[0,0]==x[0,1])
        check_Z = lambda z: len(z.shape)==2 and (z[0,0]==z[1,0]) and (z[0,0]!=z[0,1])
        
        # corners
        xc  =  self.x_corners
        zc  =  self.z_corners
        
        # assert shape
        assert check_X(xc) and check_X(xi) and (xc.shape == xi.shape)
        assert check_Z(zc) and check_Z(zi) and (zc.shape == zi.shape)
        
        # make rolls
        roll_kx = self.__shift_closest(xc[:,0], xi[:,0], self.args.Lx)
        roll_kz = self.__shift_closest(zc[0,:], zi[0,:], self.args.Lz)
        
        return roll_kx, roll_kz
        
    def __shift_closest(self, cen, c1, L):
        # shift_closest:
        #     find closest roll to each location in "cen"
        
        # assert elements are unique (and different)
        assert cen.shape == torch.unique(cen).shape and len(cen)>1
        assert c1 .shape == torch.unique(c1 ).shape and len(c1 )>1
        
        # find closest rolls (as candidates)
        inds_k = find_roll_indexes(cen, c1)
        
        # assert only 2 options are present
        assert len(inds_k) == 2
        a,b  =  inds_k
        
        # roll coordinates to check distances
        pa  =  roll_1d(c1, a, L)
        pb  =  roll_1d(c1, b, L)
        
        a_is_closer  =  torch.unique( torch.abs(cen-pa) < torch.abs(cen-pb) )
        assert a_is_closer.numel() == 1
        if a_is_closer.item():
            return a
        else:
            return b

# ------------------------------------------------------------------------
#                  Base Class: Interp2Center
# ------------------------------------------------------------------------

class Utils_Interp2Center:
    
    def calc_all_k_roll(self):
        
        # problem:
        # determine which (kx,kz) relative indexes neighbor x_cen,z_cen (can be 2 or 3 indexes)
        
        self.all_kx_roll  =  find_roll_indexes( self.args.xp_1d_centers , self.grid.x1 )
        self.all_kz_roll  =  find_roll_indexes( self.args.zp_1d_centers , self.grid.z1 )
        
        self.fit_nx       =  len(self.all_kx_roll)-1
        self.fit_nz       =  len(self.all_kz_roll)-1

        assert all( n == {'p':2,'u':1}[self.grid.seq[ind]] for ind,n in [[0,self.fit_nx],[-1,self.fit_nz]] )

# ------------------------------------------------------------------------
#                  Mixed-Order Integration (Quadrature, Cell-Centered)
# ------------------------------------------------------------------------

class Interp2Center_Quadrature_Integration(Quadrature_Integration, Utils_Interp2Center):
    def __init__(self, grid, args):
        
        self.args, self.grid = args, grid
        
        # make x_cen, z_cen
        self.x_cen = args.xp_2d_centers
        self.z_cen = args.zp_2d_centers
        
        # make all_k_roll
        self.calc_all_k_roll()
        
        # this will call quadrature integraiton
        super().__init__(grid, args)

# ------------------------------------------------------------------------
#                  Mixed-Order Integration (Midpoint, Cell-Centered)
# ------------------------------------------------------------------------

class Interp2Center_Midpoint_Integration(Base_Grid_Integration_Class, Utils_Interp2Center):
    def __init__(self, grid, args):
        
        self.args, self.grid = args, grid
        
        # make x_cen, z_cen
        self.x_cen = args.xp_2d_centers
        self.z_cen = args.zp_2d_centers
        
        # make all_k_roll
        self.calc_all_k_roll()
        
        # this __init__ method only defines a few global variables (like is_bottom, etc.)
        super().__init__(grid, args)
        
        # make n, y_cen
        self.mk_n_ycen(args.sheight)
        
        # make dxdz_2d ( as a side-product of get_XZ_amp() )
        _, _, self.dxdz_2d = get_XZamp_dxdz(self.x_cen, self.z_cen, args.Lx, args.Lz)
        
        # build fits:
        b_fit             =  self.make_b_fit()
        self.assert_almost(b_fit)
                
        # function to "get values" in main runner
        def get_values(j, K, mode):
            # f(x,z) =                   nj(x,z) * K  * b_fit(     x    ,      h(x,z) ,      z    , mode)
            f        =             (self.n[j]    * K) * b_fit(self.x_cen, self.y_cen  , self.z_cen, mode)
            return  f * self.dxdz_2d
        self.get_values = get_values
        
        self.run()
    
    def mk_n_ycen(self, sheight):
        
        # call height function
        h, grad_x, grad_z = grad_h(self.x_cen, self.z_cen, lambda x,z: sheight.get_h(x,z,side='bottom', lib=torch))
        
        # calculate normal (with top/bottom shift) (keep n[1]=1 scaling)
        self.n   =  [-grad_x, 1., -grad_z]
        
        if self.is_bottom:
            self.y_cen   =  h
        else:
            self.y_cen   =  sheight.Ly - h
            # the normal was calculated using the analytical formula
            # therefore, only the component n[1] must be reversed (not the entire vector)
            self.n[1]   *=  -1
    