# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  torch
from    math     import  pi       as  math_PI
from    os.path  import  basename
from    os.path  import  join     as  pjoin

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.utils      import  import_json, as_namedtuple

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

sin             =  torch.sin
cos             =  torch.cos
as_item         =  lambda x: x.item()

# ------------------------------------------------------------------------
#                  Main Class
# ------------------------------------------------------------------------

class HeightFunction:
    def __init__(self, fname_surf ):
        
        # read surface data
        self.fname_surf          =  fname_surf
        
        self.surf_data  =  data  =  as_namedtuple( import_json(self.fname_surf) , 'surf_json' )
        
        # import variables in surface json
        for key,val in data._asdict().items():
            setattr(self, key, val )
        
        # check that the most important variables were loaded
        assert  all( hasattr(self,k) for k in 'Lx,Ly,Lz,mnab,kx_kz_a_b,size_check'.split(',') )
        
        # verify shapes
        shape = (self.size_check,4)
        assert all( (torch.tensor(arr,dtype=torch.double).shape == shape) for arr in [self.mnab,self.kx_kz_a_b] )
        
        # verify elements of kx_kz_a_b
        _2PI    =  float(2*math_PI)
        Lx, Lz  =  self.Lx, self.Lz
        almost  =  lambda v1,v2: abs(v1-v2) < 1e-10
        for i in range(self.size_check):
            m , n , a0, b0  =  self.mnab     [i]
            kx, kz, a , b   =  self.kx_kz_a_b[i]
            
            assert almost(a0,a) and almost(b0,b)
            assert almost(kx, _2PI*m /Lx) and almost(kz, _2PI*n /Lz)
    
    def get_h(self, X, Z):
        # lib as a parameter opens to possibility to call Pytorch-Cuda (if needed)
        h          =  0.
        
        for kx, kz, a, b  in  self.kx_kz_a_b:
            theta  =  kx*X + kz*Z
            h     +=  a*cos(theta) + b*sin(theta)
        return h
    
    def grad_h(self, X, Z, cos=cos, sin=sin, dict_output=False):
        
        h               =  0.
        grad_x, grad_z  =  0.      , 0.
        
        for kx, kz, a, b  in  self.kx_kz_a_b:
            theta    =  kx*X + kz*Z
            cos_o    =  cos(theta)
            sin_o    =  sin(theta)
            h       +=  a*cos_o    + b*sin_o
            
            common   =  (-a)*sin_o + b*cos_o
            grad_x  +=  kx * common
            grad_z  +=  kz * common
        
        if dict_output:
            return dict(h=h, grad_x=grad_x, grad_z=grad_z)
        else:
            return h, grad_x, grad_z