# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  torch
from    collections  import namedtuple

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.utils           import  lmap
from  core.misc_tools      import  q2d_integral, Write_Tec2d
from  core.shape_function  import  ShapeFunction

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

ones_like  =  torch.ones_like

# ------------------------------------------------------------------------
#                  Main Hex Integration
# ------------------------------------------------------------------------

class Hex_Integration_Grid:
    def __init__(self, grid, args):
        
        self.grid  =  grid
        self.args  =  args
        
        # fetch Lmode
        Lmode      =  args.Lmode
        
        # detect side
        side       =  args.side
        assert side in ('top','bottom')
        self.is_bottom = side == 'bottom'
        
        # build fits
        h_fit      =  self.make_h_fit()
        b_fit      =  self.make_b_fit()
        
        # setup variables
        h          =  lambda x, z, mode:  h_fit(x, 1., z, mode)
        x_avg      =  h_fit.X_avg
        x_amp      =  h_fit.X_amp
        z_avg      =  h_fit.Z_avg
        z_amp      =  h_fit.Z_amp
        
        nx, nz     =  2, 2 # polynomials have degree 2 in total
        
        # flipn      =  1 if self.is_bottom else -1
        
        # at the bottom side, the normal has the right orientation
        # however, at the top, the normal is calculated with mirrored points,
        # and it points inwards. therefore, the ENTIRE vector must be reversed.
        
        # note: when we calculated the normal through the analytical function,
        #       only the n[1] component was reversed (unlike here), because
        #       the reference calculation  points were also at the bottom surface
        
        if self.is_bottom:
            
            n = [  lambda x,z:  -h(x, z, 'dx')  ,
                   lambda x,z:   1.0            ,
                   lambda x,z:  -h(x, z, 'dz')  ]
        
        else:
            # reversed formula, top side
            n = [  lambda x,z:   h(x, z, 'dx')  ,
                   lambda x,z:  -1.0            ,
                   lambda x,z:   h(x, z, 'dz')  ]
        
        all_tec_2d    =  [ x_avg,  h(x_avg, z_avg, 'f') , z_avg ]
        all_tec_tags  =  [ 'X'  ,  'Y'                  , 'Z'   ]
        
        self.Dsum  =  Dsum  =  [{},{},{}]
        
        for i in range(3):
            for j in range(3):
                if Lmode[i][j]:
                    t        =  Lmode[i][j]
                    K, mode  =  t.K, t.mode
                    nj       =  n[j]
                    
                    # f(x,z) =                      nj(x,z) * K * b    (x, h(x,z)     ,z, mode)
                    f                =  lambda x,z: nj(x,z) * K * b_fit(x, h(x,z,'f'), z, mode)
                    values           =  q2d_integral(f, x_avg, x_amp, nx, z_avg, z_amp, nz)
                    
                    assert not t.tag_sum in Dsum[i]
                    Dsum[i][t.tag_sum]  =  values.sum().item()
                    
                    if t.plot:
                        all_tec_tags.append(t.tag_tec)
                        all_tec_2d  .append(values)
        
        if len(all_tec_2d) > 3:
            Write_Tec2d(all_tec_tags, all_tec_2d, args.tec2dname)
    
    def make_h_fit(self):
        
        grid  =  self.grid
        
        # bilinear element (x,z):
        nx = nz = 1
        # y direction is constant:
        ny      = 0
        
        all_X   = self.gather_quad( grid.x_2d , L0 = self.args.Lx)
        all_Y   = lmap( ones_like, all_X )
        all_Z   = self.gather_quad( grid.z_2d , L1 = self.args.Lz)
        all_U   = self.gather_quad( grid.y_bottom if  self.is_bottom  else grid.y_top )
        
        return ShapeFunction(all_X, all_Y, all_Z, all_U, nx, ny, nz)
    
    
    def make_b_fit(self):
        grid         =  self.grid
        gather_quad  =  self.gather_quad
        
        get_side     =  lambda tag: getattr(grid, f'{tag}_{self.args.side}')
        
        # bilinear shape function in xz (ny may vary)
        nx = nz = 1
        
        
        all_Y   = []
        all_U   = []
        
        # interior layer of values
        i_side     =  get_side('i'   )
        vals_side  =  get_side('vals')
        
        layers     = vals_side.shape[-1]
        
        # check whether to add walls, or more interior points
        assert grid.tag in 'UVWPT'
        if grid.tag == 'P':
            assert layers == 2
            ny = 1
        else:
            # wall velocities
            all_Y   +=  gather_quad( get_side('y') )
            all_U   +=  lmap(torch.zeros_like, all_Y)
            ny       =  layers
        
        assert all( (len(arr.shape)==3 and (arr.shape[-1] == layers)) for arr in [i_side,vals_side] )
        
        # base 2d grid
        all_X  =  gather_quad( grid.x_2d , L0 = self.args.Lx) * (ny+1)
        all_Z  =  gather_quad( grid.z_2d , L1 = self.args.Lz) * (ny+1)
        
        for pos in range(layers):
            all_Y  +=  gather_quad( grid.y1[ i_side   [:,:,pos] ] )
            all_U  +=  gather_quad(          vals_side[:,:,pos]   )
        
        return ShapeFunction(all_X, all_Y, all_Z, all_U, nx, ny, nz)
    
    def gather_quad(self, p, L0 = None, L1 = None):
        p  =  torch.clone(p)
        assert len(p.shape) == 2
        
        def roll(x,a):
            x = torch.roll(torch.clone(x), -1, dims = a)
            if   (a == 0) and L0:
                x[-1,:] += L0
            elif (a == 1) and L1:
                x[:,-1] += L1
            return x
        
        p0, p1  =  lmap(lambda i: roll(p,i), range(2))
        p01     =  roll(p0,1)
        return [ p, p0, p1, p01 ]

# alternative implementation (more verbose, without np.roll)
# (periodicity must be incorporated before using)
#   def gather_quad(self, P):
#       P = np.copy(P)
#       assert len(P.shape) == 2
#       a,b              =  P.shape
#       P_per            =  np.zeros((a+1,b+1),dtype = float) * np.nan
#       P_per[ :a, :b]   = P
#       P_per[-1  ,  : ] = P_per[0,:]
#       P_per[ :  , -1 ] = P_per[:,0]
#       return [ P_per[:a,:b], P_per[1:,:b], P_per[:a,1:], P_per[1:,1:] ]
#       # assert all( (not (a1-a2).any()) for a1,a2 in zip(gather_quad(A), gather_quad2(A) ) )
