# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  torch
from    collections  import  namedtuple

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.utils           import  lmap
from  core.misc_tools      import  q2d_integral, Write_Tec2d
from  core.shape_function  import  ShapeFunction

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

ones_like  =  torch.ones_like

# ------------------------------------------------------------------------
#                  Roll Function
# ------------------------------------------------------------------------


def roll(x, k, a, L0,L1):
    
    if k:
        
        x = torch.roll(torch.clone(x), k, dims = a)
        
        if   (a == 0) and L0:
            if k<0:
                # array (x) was pulled "towards the origin"
                # therefore, the last x[-|k|:,:] positions must be
                # shifted by L0
                x[k:,:] += L0
            else:
                # opposite scenario:
                # array (x) was pulled "towards the end"
                # the fist "k" elements must be substracted L0
                x[:k,:] -= L0
        
        elif (a == 1) and L1:
            # similar cases as before, but for the inner dimension
            if k<0:
                x[:,k:] += L1
            else:
                x[:,:k] -= L1
    return x

# ------------------------------------------------------------------------
#                  Main Hex Integration
# ------------------------------------------------------------------------

class Interp2Center_Grid:
    def __init__(self, grid, args):
        
        # register input
        self.grid  =  grid
        self.args  =  args
        
        # fetch Lmode
        Lmode      =  args.Lmode
        
        # detect side
        side       =  args.side
        assert side in ('top','bottom')
        self.is_bottom = side == 'bottom'
        
        # build fit
        b_fit      =  self.make_b_fit()
        
        # fetch normal
        n          =  self.calc_n()
        
        # initialize tec2d
        all_tec_2d    =  [ x_cen,  y_cen , z_cen ]
        all_tec_tags  =  [ 'X'  ,  'Y'   , 'Z'   ]
        
        # mk Dsum
        self.Dsum  =  Dsum  =  [{},{},{}]
        
        for i in range(3):
            for j in range(3):
                if Lmode[i][j]:
                    t        =  Lmode[i][j]
                    K, mode  =  t.K, t.mode
                    
                    # f(x,z) =           nj(x,z) * K  * b    (x    , h(x,z) , z    , mode)
                    f                =  (n[j]    * K) * b_fit(x_cen, y_cen  , z_cen, mode)
                    values           =  f * dxdz_2d
                    
                    assert not t.tag_sum in Dsum[i]
                    Dsum[i][t.tag_sum]  =  values.sum().item()
                    
                    if t.plot:
                        all_tec_tags.append(t.tag_tec)
                        all_tec_2d  .append(values)
        
        if len(all_tec_2d) > 3:
            Write_Tec2d(all_tec_tags, all_tec_2d, args.tec2dname)
    
    def make_b_fit(self):
        grid         =  self.grid
        seq          =  grid.seq
        
        gather_2d  =  self.gather_2d
        
        # get_side: fetch grid variables, such as vals_bottom, vals_top, etc.
        get_side     =  lambda tag: getattr(grid, f'{tag}_{self.args.side}')
        
        # bilinear shape function in xz (ny may vary)
        # nx = nz = 1
        
        
        all_Y   = []
        all_U   = []
        
        # interior layer of values
        i_side     =  get_side('i'   )
        vals_side  =  get_side('vals')
        
        # note: "layers" corresponds to the former "n_side" parameter
        layers     = vals_side.shape[-1]
        
        # check whether to add walls, or only interior points
        assert grid.tag in 'UVWPT'
        if grid.tag == 'P':
            assert layers == 2
            ny = 1
        else:
            # wall velocities
            all_Y   +=  gather_2d( get_side('y') )
            all_U   +=  lmap(torch.zeros_like, all_Y)
            ny       =  layers
        
        assert all( (len(arr.shape)==3 and (arr.shape[-1] == layers)) for arr in [i_side,vals_side] )
        
        # base 2d grid
        all_X  =  gather_2d( grid.x_2d , L0 = self.args.Lx) * (ny+1)
        all_Z  =  gather_2d( grid.z_2d , L1 = self.args.Lz) * (ny+1)
        
        for pos in range(layers):
            all_Y  +=  gather_2d( grid.y1[ i_side   [:,:,pos] ] )
            all_U  +=  gather_2d(          vals_side[:,:,pos]   )
        
        return ShapeFunction(all_X, all_Y, all_Z, all_U, nx, ny, nz)
    
    def gather_2d(self, p, L0 = None, L1 = None):
        
        assert len(p.shape) == 2
        
        result  = []
        
        for     k0 in self.all_k0_roll:
            
            x = roll(p, k0, 0, L0,L1)
            
            for k1 in self.all_k1_roll:
                
                result += [ roll(x, k1, 1 ,L0,L1) ]
        
        return result
